###########################################################################
# MumbleIce is Copyright (C) 2021 Kyle Robbertze <kyle@bitcast.co.za>
#
# MumbleIce is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, or
# any later version as published by the Free Software Foundation.
#
# MumbleIce is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with MumbleIce. If not, see <http://www.gnu.org/licenses/>.
###########################################################################
import logging
import time
from importlib.resources import files, as_file
from .mumble import MumbleConnector
from .icecast import IcecastConnector
from .utils import Watchdog, SilenceError, ConfigurationError, parse_message, BUFFER_DURATION, WATCHDOG_RATE, NUM_CHANNELS


class Bot:
    def __init__(self, mumble, icecast, command_prefix, autoconnect=False):
        if BUFFER_DURATION % 10 != 0:
            raise ConfigurationError('BUFFER_DURATION must be 10ms or a multiple thereof')
        if NUM_CHANNELS not in [1, 2]:
            raise ConfigurationError('NUM_CHANNELS must be 1 (mono) or (stereo). Other values are not supported')
        self.mumble = mumble
        self.mumble.set_message_callback(self.mumble_message)
        self.icecast = icecast
        self.logger = logging.getLogger(__name__)
        self.command_prefix = command_prefix
        self.commands = {
            'connect': self.connect_icecast,
            'disconnect': self.disconnect_icecast,
            'status': self.show_icecast_status,
        }
        self.autoconnect = autoconnect
        self.timer = Watchdog(WATCHDOG_RATE/1000, self.write_audio)

    def run(self):
        self.logger.debug('Starting MumbleIce bot')
        self.logger.debug('Connecting to Mumble...')
        self.mumble.start()
        self.logger.info('Connected to Mumble')
        self.logger.info('Started MumbleIce bot')

        if self.autoconnect:
            self.connect_icecast()

        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            self.timer.stop()
            self.icecast.stop()
            self.mumble.stop()

    def mumble_message(self, message):
        self.logger.info(f'Message received: {message.message}')
        msg = parse_message(message.message)
        if msg.startswith(self.command_prefix) and len(msg) > len(self.command_prefix):
            msg = msg[len(self.command_prefix):]
            msg = msg.split(' ')
            cmd = self.commands.get(msg[0])
            if cmd == None:
                self.logger.debug(f'Not recognized command: {message.message}')
                self.mumble.send_message('Command not recognised')
                return
            cmd()

    def connect_icecast(self):
        if self.icecast.is_connected:
            self.mumble.send_message('Icecast already connected')
        else:
            self.icecast.start()
            self.mumble.set_get_sound(True)
            self.mumble.send_message('Icecast stream started')
            audio_file = files('mumbleice').joinpath('resources').joinpath('streaming_started.wav')
            with as_file(audio_file) as audio:
                self.mumble.send_audio(audio.resolve())
            self.timer.start()

    def disconnect_icecast(self):
        self.timer.stop()
        if self.icecast.is_connected:
            self.icecast.stop()
            self.mumble.send_message('Icecast streaming stopped')
            audio_file = files('mumbleice').joinpath('resources').joinpath('streaming_stopped.wav')
            with as_file(audio_file) as audio:
                self.mumble.send_audio(audio.resolve())
            self.mumble.set_get_sound(False)
            # Ensure that the timer has stopped and will not restart
            self.timer.stop()
        else:
            self.mumble.send_message('Icecast already disconnected')

    def write_audio(self):
        try:
            audio = self.mumble.get_audio(BUFFER_DURATION)
            self.icecast.write(audio)
            self.timer.reset()
        except SilenceError:
            self.logger.warning(f'No audio received from Mumble for the last {self.mumble.max_silence} s. Disconnecting from Icecast...')
            self.mumble.send_message(f'No audio received for the last {self.mumble.max_silence} s. Disconnecting from Icecast...')
            self.disconnect_icecast()
        except BrokenPipeError:
            self.mumble.send_message('Icecast disconnected unexpectedly. Streaming stopped')
            self.disconnect_icecast()

    def show_icecast_status(self):
        if self.icecast.is_connected:
            self.mumble.send_message('Icecast is <b>connected</b> and streaming')
        else:
            self.mumble.send_message('Icecast is <b>disconnected</b> and not streaming')



###########################################################################
# MumbleIce is Copyright (C) 2021 Kyle Robbertze <kyle@bitcast.co.za>
#
# MumbleIce is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, or
# any later version as published by the Free Software Foundation.
#
# MumbleIce is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with MumbleIce. If not, see <http://www.gnu.org/licenses/>.
###########################################################################
import logging
import ffmpeg
import subprocess
from .utils import SAMPLES_PER_SECOND, NUM_CHANNELS


class IcecastConnector:
    def __init__(self, server, port, username, password, mount_point):
        self.logger = logging.getLogger(__name__)
        self.icecast_string = f'icecast://{username}:{password}@{server}:{port}{mount_point}'
        self.icecast_stream = None
        self.is_connected = False

    def start(self):
        args = (
            ffmpeg
            .input('pipe:', format='s16le', ar=SAMPLES_PER_SECOND, ac=NUM_CHANNELS)
            .output(self.icecast_string,
                    codec="libmp3lame",
                    f='mp3',
                    ac=2,
                    legacy_icecast=1,
                    sample_fmt='fltp',
                    content_type='audio/mpeg',
                    **{
                        'b:a': '132k',
                    })
            .global_args('-hide_banner')
            .compile()
        )
        args.insert(1, '-re')
        self.icecast_stream = subprocess.Popen(args, stdin=subprocess.PIPE)
        self.logger.info('Icecast stream started')
        self.is_connected = True

    def write(self, pcm):
        if self.icecast_stream.poll():
            self.logger.warning('Icecast stream disconnected unexpectedly, reconnecting...')
            self.is_connected = False
            self.start()
        self.logger.debug('Writing data to FFMpeg')
        try:
            self.icecast_stream.stdin.write(pcm)
        except ValueError:
            self.logger.debug('Attempted to write to closed Icecast stream')
            self.is_connected = False
        except BrokenPipeError as e:
            self.logger.warning('Icecast disappeared unexpectedly')
            self.disconnect_icecast()
            raise e

    def stop(self):
        if self.icecast_stream:
            self.icecast_stream.stdin.close()
        self.logger.info('Disconnected from Icecast')
        self.is_connected = False



###########################################################################
# MumbleIce is Copyright (C) 2021 Kyle Robbertze <kyle@bitcast.co.za>
#
# MumbleIce is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, or
# any later version as published by the Free Software Foundation.
#
# MumbleIce is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with MumbleIce. If not, see <http://www.gnu.org/licenses/>.
###########################################################################



###########################################################################
# MumbleIce is Copyright (C) 2021 Kyle Robbertze <kyle@bitcast.co.za>
#
# MumbleIce is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, or
# any later version as published by the Free Software Foundation.
#
# MumbleIce is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with MumbleIce. If not, see <http://www.gnu.org/licenses/>.
###########################################################################
import logging
import pymumble_py3 as pymumble
from pydub import AudioSegment
from .utils import SilenceError, SAMPLES_PER_SECOND, NUM_CHANNELS, MAX_SILENCE_DURATION, read_file_buffer
import subprocess


class MumbleConnector:
    def __init__(
        self,
        server,
        port,
        username,
        password,
        channel,
        max_silence=MAX_SILENCE_DURATION,
    ):
        self.server = server
        self.port = port
        self.username = username
        self.password = password
        self.channel = channel
        self.mumble = pymumble.Mumble(server, username,
                                      port, password,
                                      reconnect=True)
        self.logger = logging.getLogger(__name__)
        self.silence_count = 0
        self.max_silence = max_silence * 1000 # ms

    def start(self):
        self.mumble.start()
        self.mumble.is_ready()
        #TODO: Crashes with RuntimeError("threads can only be started once")
        self.mumble.callbacks.set_callback(
            pymumble.constants.PYMUMBLE_CLBK_DISCONNECTED,
            self.restart)
        self.logger.info('Mumble client started')
        try:
            channel = self.mumble.channels.find_by_name(self.channel)
            channel.move_in()
        except pymumble.errors.UnknownChannelError:
            self.logger.error(f'Unable to find channel {self.channel}')

    def restart(self):
        self.mumble = pymumble.Mumble(self.server, self.username,
                                      self.port, self.password,
                                      reconnect=True)
        self.silence_count = 0
        self.start()

    def stop(self):
        self.mumble.callbacks.remove_callback(
            pymumble.constants.PYMUMBLE_CLBK_DISCONNECTED,
            self.restart)
        self.mumble.stop()
        self.logger.info('Disconnected from Mumble')

    def set_message_callback(self, fn):
        self.mumble.callbacks.set_callback(
            pymumble.constants.PYMUMBLE_CLBK_TEXTMESSAGERECEIVED,
            fn)

    def set_get_sound(self, receive):
        self.logger.debug('Reset silence count')
        self.silence_count = 0
        self.mumble.set_receive_sound(receive)

    def send_message(self, message):
        self.logger.debug(f'Sending message {message} to mumble channel')
        self.mumble.my_channel().send_text_message(message)

    def send_audio(self, path):
        with read_file_buffer(path, 1024) as buffer:
            for audio in buffer:
                self.mumble.sound_output.add_sound(audio)

    def get_audio(self, buffer_size):
        self.logger.debug('Fetching mumble audio')
        audio_length = buffer_size/1000
        audio = AudioSegment.silent(duration=buffer_size, frame_rate=SAMPLES_PER_SECOND)
        silence = True
        for session_id, user in self.mumble.users.items():
            user_audio = user.sound
            if user_audio.is_sound():
                silence = False
                audio_data = AudioSegment(
                    data=user_audio.get_sound(audio_length).pcm,
                    sample_width=2, # bytes => 16 bit
                    frame_rate=SAMPLES_PER_SECOND,
                    channels=1
                )
                audio = audio.overlay(audio_data)
        if NUM_CHANNELS == 2:
            audio = AudioSegment.from_mono_audiosegments(audio, audio)

        if self.max_silence > 0:
            if silence:
                self.logger.debug(f'Silence detected from Mumble for {buffer_size}ms')
                self.silence_count = self.silence_count + buffer_size
            else:
                self.silence_count = 0
            if self.silence_count > self.max_silence:
                self.silence_count = 0
                raise SilenceError()
        return audio.raw_data
